package com.sun.org.apache.bcel.internal.generic;

import java.util.Objects;
import java.util.Iterator;
import com.sun.org.apache.bcel.internal.classfile.ConstantPool;
import com.sun.org.apache.bcel.internal.classfile.AnnotationEntry;
import com.sun.org.apache.bcel.internal.classfile.RuntimeInvisibleAnnotations;
import com.sun.org.apache.bcel.internal.classfile.RuntimeVisibleAnnotations;
import com.sun.org.apache.bcel.internal.classfile.Annotations;
import com.sun.org.apache.bcel.internal.classfile.JavaClass;
import com.sun.org.apache.bcel.internal.classfile.SourceFile;
import java.util.ArrayList;
import com.sun.org.apache.bcel.internal.util.BCELComparator;
import com.sun.org.apache.bcel.internal.classfile.Attribute;
import com.sun.org.apache.bcel.internal.classfile.Method;
import com.sun.org.apache.bcel.internal.classfile.Field;
import java.util.List;
import com.sun.org.apache.bcel.internal.classfile.AccessFlags;

public class ClassGen extends AccessFlags implements Cloneable
{
    private String class_name;
    private String super_class_name;
    private final String file_name;
    private int class_name_index;
    private int superclass_name_index;
    private int major;
    private int minor;
    private ConstantPoolGen cp;
    private final List<Field> field_vec;
    private final List<Method> method_vec;
    private final List<Attribute> attribute_vec;
    private final List<String> interface_vec;
    private final List<AnnotationEntryGen> annotation_vec;
    private static BCELComparator _cmp;
    private List<ClassObserver> observers;
    
    public ClassGen(String class_name, String super_class_name, String file_name, int access_flags, String[] interfaces, ConstantPoolGen cp) {
        super(access_flags);
        this.class_name_index = -1;
        this.superclass_name_index = -1;
        this.major = 45;
        this.minor = 3;
        this.field_vec = new ArrayList();
        this.method_vec = new ArrayList();
        this.attribute_vec = new ArrayList();
        this.interface_vec = new ArrayList();
        this.annotation_vec = new ArrayList();
        this.class_name = class_name;
        this.super_class_name = super_class_name;
        this.file_name = file_name;
        this.cp = cp;
        if (file_name != null) {
            this.addAttribute(new SourceFile(cp.addUtf8("SourceFile"), 2, cp.addUtf8(file_name), cp.getConstantPool()));
        }
        this.class_name_index = cp.addClass(class_name);
        this.superclass_name_index = cp.addClass(super_class_name);
        if (interfaces != null) {
            for (String interface1 : interfaces) {
                this.addInterface(interface1);
            }
        }
    }
    
    public ClassGen(String class_name, String super_class_name, String file_name, int access_flags, String[] interfaces) {
        this(class_name, super_class_name, file_name, access_flags, interfaces, new ConstantPoolGen());
    }
    
    public ClassGen(JavaClass clazz) {
        super(clazz.getAccessFlags());
        this.class_name_index = -1;
        this.superclass_name_index = -1;
        this.major = 45;
        this.minor = 3;
        this.field_vec = new ArrayList();
        this.method_vec = new ArrayList();
        this.attribute_vec = new ArrayList();
        this.interface_vec = new ArrayList();
        this.annotation_vec = new ArrayList();
        this.class_name_index = clazz.getClassNameIndex();
        this.superclass_name_index = clazz.getSuperclassNameIndex();
        this.class_name = clazz.getClassName();
        this.super_class_name = clazz.getSuperclassName();
        this.file_name = clazz.getSourceFileName();
        this.cp = new ConstantPoolGen(clazz.getConstantPool());
        this.major = clazz.getMajor();
        this.minor = clazz.getMinor();
        Attribute[] attributes = clazz.getAttributes();
        AnnotationEntryGen[] annotations = this.unpackAnnotations(attributes);
        Method[] methods = clazz.getMethods();
        Field[] fields = clazz.getFields();
        String[] interfaces;
        for (String interface1 : interfaces = clazz.getInterfaceNames()) {
            this.addInterface(interface1);
        }
        for (Attribute attribute : attributes) {
            if (!(attribute instanceof Annotations)) {
                this.addAttribute(attribute);
            }
        }
        for (AnnotationEntryGen annotation : annotations) {
            this.addAnnotationEntry(annotation);
        }
        for (Method method : methods) {
            this.addMethod(method);
        }
        for (Field field : fields) {
            this.addField(field);
        }
    }
    
    private AnnotationEntryGen[] unpackAnnotations(Attribute[] attrs) {
        List<AnnotationEntryGen> annotationGenObjs = new ArrayList();
        for (Attribute attr : attrs) {
            if ((attr instanceof RuntimeVisibleAnnotations)) {
                RuntimeVisibleAnnotations rva = (RuntimeVisibleAnnotations)attr;
                AnnotationEntry[] annos;
                for (AnnotationEntry a : annos = rva.getAnnotationEntries()) {
                    annotationGenObjs.add(new AnnotationEntryGen(a, this.getConstantPool(), false));
                }
            }
            else if ((attr instanceof RuntimeInvisibleAnnotations)) {
                RuntimeInvisibleAnnotations ria = (RuntimeInvisibleAnnotations)attr;
                AnnotationEntry[] annos;
                for (AnnotationEntry a : annos = ria.getAnnotationEntries()) {
                    annotationGenObjs.add(new AnnotationEntryGen(a, this.getConstantPool(), false));
                }
            }
        }
        return (AnnotationEntryGen[])annotationGenObjs.toArray(new AnnotationEntryGen[annotationGenObjs.size()]);
    }
    
    public JavaClass getJavaClass() {
        int[] interfaces = this.getInterfaces();
        Field[] fields = this.getFields();
        Method[] methods = this.getMethods();
        Attribute[] attributes = null;
        if (this.annotation_vec.isEmpty()) {
            attributes = this.getAttributes();
        }
        else {
            Attribute[] annAttributes = AnnotationEntryGen.getAnnotationAttributes(this.cp, this.getAnnotationEntries());
            attributes = new Attribute[this.attribute_vec.size() + annAttributes.length];
            this.attribute_vec.toArray(attributes);
            System.arraycopy(annAttributes, 0, attributes, this.attribute_vec.size(), annAttributes.length);
        }
        ConstantPool _cp = this.cp.getFinalConstantPool();
        return new JavaClass(this.class_name_index, this.superclass_name_index, this.file_name, this.major, this.minor, super.getAccessFlags(), _cp, interfaces, fields, methods, attributes);
    }
    
    public void addInterface(String name) {
        this.interface_vec.add(name);
    }
    
    public void removeInterface(String name) {
        this.interface_vec.remove(name);
    }
    
    public int getMajor() {
        return this.major;
    }
    
    public void setMajor(int major) {
        this.major = major;
    }
    
    public void setMinor(int minor) {
        this.minor = minor;
    }
    
    public int getMinor() {
        return this.minor;
    }
    
    public void addAttribute(Attribute a) {
        this.attribute_vec.add(a);
    }
    
    public void addAnnotationEntry(AnnotationEntryGen a) {
        this.annotation_vec.add(a);
    }
    
    public void addMethod(Method m) {
        this.method_vec.add(m);
    }
    
    public void addEmptyConstructor(int access_flags) {
        InstructionList il = new InstructionList();
        il.append(InstructionConst.THIS);
        il.append(new INVOKESPECIAL(this.cp.addMethodref(this.super_class_name, "<init>", "()V")));
        il.append(InstructionConst.RETURN);
        MethodGen mg = new MethodGen(access_flags, Type.VOID, Type.NO_ARGS, null, "<init>", this.class_name, il, this.cp);
        mg.setMaxStack(1);
        this.addMethod(mg.getMethod());
    }
    
    public void addField(Field f) {
        this.field_vec.add(f);
    }
    
    public boolean containsField(Field f) {
        return this.field_vec.contains(f);
    }
    
    public Field containsField(String name) {
        for (Field f : this.field_vec) {
            if (f.getName().equals(name)) {
                return f;
            }
            continue;
        }
        return null;
    }
    
    public Method containsMethod(String name, String signature) {
        for (Method m : this.method_vec) {
            if (m.getName().equals(name) && m.getSignature().equals(signature)) {
                return m;
            }
            continue;
        }
        return null;
    }
    
    public void removeAttribute(Attribute a) {
        this.attribute_vec.remove(a);
    }
    
    public void removeMethod(Method m) {
        this.method_vec.remove(m);
    }
    
    public void replaceMethod(Method old, Method new_) {
        if (new_ == null) {
            throw new ClassGenException("Replacement method must not be null");
        }
        int i = this.method_vec.indexOf(old);
        if (i < 0) {
            this.method_vec.add(new_);
        }
        else {
            this.method_vec.set(i, new_);
        }
        return;
    }
    
    public void replaceField(Field old, Field new_) {
        if (new_ == null) {
            throw new ClassGenException("Replacement method must not be null");
        }
        int i = this.field_vec.indexOf(old);
        if (i < 0) {
            this.field_vec.add(new_);
        }
        else {
            this.field_vec.set(i, new_);
        }
        return;
    }
    
    public void removeField(Field f) {
        this.field_vec.remove(f);
    }
    
    public String getClassName() {
        return this.class_name;
    }
    
    public String getSuperclassName() {
        return this.super_class_name;
    }
    
    public String getFileName() {
        return this.file_name;
    }
    
    public void setClassName(String name) {
        this.class_name = name.replace('/', '.');
        this.class_name_index = this.cp.addClass(name);
    }
    
    public void setSuperclassName(String name) {
        this.super_class_name = name.replace('/', '.');
        this.superclass_name_index = this.cp.addClass(name);
    }
    
    public Method[] getMethods() {
        return (Method[])this.method_vec.toArray(new Method[this.method_vec.size()]);
    }
    
    public void setMethods(Method[] methods) {
        this.method_vec.clear();
        for (Method method : methods) {
            this.addMethod(method);
        }
    }
    
    public void setMethodAt(Method method, int pos) {
        this.method_vec.set(pos, method);
    }
    
    public Method getMethodAt(int pos) {
        return (Method)this.method_vec.get(pos);
    }
    
    public String[] getInterfaceNames() {
        int size = this.interface_vec.size();
        String[] interfaces = new String[size];
        this.interface_vec.toArray(interfaces);
        return interfaces;
    }
    
    public int[] getInterfaces() {
        int size = this.interface_vec.size();
        int[] interfaces = new int[size];
        for (int i = 0; i < size; ++i) {
            interfaces[i] = this.cp.addClass((String)this.interface_vec.get(i));
        }
        return interfaces;
    }
    
    public Field[] getFields() {
        return (Field[])this.field_vec.toArray(new Field[this.field_vec.size()]);
    }
    
    public Attribute[] getAttributes() {
        return (Attribute[])this.attribute_vec.toArray(new Attribute[this.attribute_vec.size()]);
    }
    
    public AnnotationEntryGen[] getAnnotationEntries() {
        return (AnnotationEntryGen[])this.annotation_vec.toArray(new AnnotationEntryGen[this.annotation_vec.size()]);
    }
    
    public ConstantPoolGen getConstantPool() {
        return this.cp;
    }
    
    public void setConstantPool(ConstantPoolGen constant_pool) {
        this.cp = constant_pool;
    }
    
    public void setClassNameIndex(int class_name_index) {
        this.class_name_index = class_name_index;
        this.class_name = this.cp.getConstantPool().getConstantString(class_name_index, (byte)7).replace('/', '.');
    }
    
    public void setSuperclassNameIndex(int superclass_name_index) {
        this.superclass_name_index = superclass_name_index;
        this.super_class_name = this.cp.getConstantPool().getConstantString(superclass_name_index, (byte)7).replace('/', '.');
    }
    
    public int getSuperclassNameIndex() {
        return this.superclass_name_index;
    }
    
    public int getClassNameIndex() {
        return this.class_name_index;
    }
    
    public void addObserver(ClassObserver o) {
        if (this.observers == null) {
            this.observers = new ArrayList();
        }
        this.observers.add(o);
    }
    
    public void removeObserver(ClassObserver o) {
        if (this.observers != null) {
            this.observers.remove(o);
        }
    }
    
    public void update() {
        if (this.observers != null) {
            for (ClassObserver observer : this.observers) {
                observer.notify(this);
            }
        }
    }
    
    public Object clone() {
        try {
            return super.clone();
        }
        catch (CloneNotSupportedException e) {
            throw new Error("Clone Not Supported");
        }
    }
    
    public static BCELComparator getComparator() {
        return ClassGen._cmp;
    }
    
    public static void setComparator(BCELComparator comparator) {
        ClassGen._cmp = comparator;
    }
    
    @Override
    public boolean equals(Object obj) {
        return ClassGen._cmp.equals(this, obj);
    }
    
    @Override
    public int hashCode() {
        return ClassGen._cmp.hashCode(this);
    }
    
    static {
        ClassGen._cmp = new BCELComparator() {
            @Override
            public boolean equals(Object o1, Object o2) {
                ClassGen THIS = (ClassGen)o1;
                ClassGen THAT = (ClassGen)o2;
                return Objects.equals(THIS.getClassName(), THAT.getClassName());
            }
            
            @Override
            public int hashCode(Object o) {
                ClassGen THIS = (ClassGen)o;
                return THIS.getClassName().hashCode();
            }
        };
    }
}
